from bindings import *
from models import models
import sys

# Test CPU features
use_simd = True

if len(sys.argv) > 1 and sys.argv[1] == '--no_simd':
    use_simd = False

cpu_check_features()

if use_simd and not cpu_enable_simd:
    raise Exception('Expected SIMD intrinsics to be enabled')

if not use_simd and cpu_enable_simd:
    raise Exception('Expected SIMD intrinsics to be disabled')

#----------------------------------------

# Test CRC
test_data = bytes(b & 0xff for b in range(300))
failed = False

for name, model in models.items():
    print(name)
    params = crc_params(*model)

    # Test 1: Test crc_table
    value = crc_table(params, params.init, b'123456789')
    print("table:           ", hex(value), hex(model.check), value == model.check)

    if value != model.check:
        failed = True

    # Test 2: Test crc_calc when len < 128
    value = crc_calc(params, params.init, b'123456789')
    print("clmul & table:   ", hex(value), hex(model.check), value == model.check)

    if value != model.check:
        failed = True

    # Test 3: Test crc_calc when len > 128
    value = crc_calc(params, params.init, test_data)
    value2 = crc_table(params, params.init, test_data)
    print("clmul:           ", hex(value), hex(value2), value == value2)

    if value != value2:
        failed = True

    # Test 4: Test crc_calc in chunks
    value = crc_calc(params, params.init, test_data[:150])
    value = crc_calc(params, value, test_data[150:])
    value2 = crc_calc(params, params.init, test_data)
    print("clmul chunks:    ", hex(value), hex(value2), value == value2)

    if value != value2:
        failed = True

    # Test 5: Test crc_calc with an unaligned buffer
    for i in range(1, 16):
        value = crc_calc_unaligned(params, params.init, test_data, i)
        value2 = crc_calc(params, params.init, test_data[i:])

        if value != value2:
            print("clmul unaligned: ", hex(value), hex(value2), value == value2)
            failed = True

    # Test 6: Test crc_combine_constant and crc_combine_fixed
    xp = crc_combine_constant(params, 4)
    value = crc_calc(params, params.init, b'12345')
    value2 = crc_calc(params, params.init, b'6789')
    value = crc_combine_fixed(params, value, value2, xp)
    print("combine constant:", hex(value), hex(model.check), value == model.check)

    if value != model.check:
        failed = True

    # Test 7: Test crc_combine
    value = crc_calc(params, params.init, b'12345')
    value2 = crc_calc(params, params.init, b'6789')
    value = crc_combine(params, value, value2, 4)
    print("combine:         ", hex(value), hex(model.check), value == model.check)

    if value != model.check:
        failed = True

    print()

if failed:
    raise Exception("Test failed")
else:
    print("The test ran successfully!")


import ctypes
import os
import sys

if sys.platform == 'win32':
    ext = '.dll'
elif sys.platform == 'darwin':
    ext = '.dylib'
else:
    ext = '.so'

_crc = ctypes.CDLL(os.path.join(os.path.dirname(__file__), 'crc' + ext))

# Note: Update this definition when the equivalent C code is changed
class params_t(ctypes.Structure):
    _fields_ = [('width', ctypes.c_uint8),
               ('poly', ctypes.c_uint64),
               ('refin', ctypes.c_bool),
               ('refout', ctypes.c_bool),
               ('init', ctypes.c_uint64),
               ('xorout', ctypes.c_uint64),
               ('k1', ctypes.c_uint64),
               ('k2', ctypes.c_uint64),
               ('table', ctypes.c_uint64 * 256),
               ('combine_table', ctypes.c_uint64 * 64)]

_crc.cpu_check_features.argtypes = []

_crc.crc_params.argtypes = [ctypes.c_uint8, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_bool, ctypes.c_bool, ctypes.c_uint64, ctypes.c_uint64, ctypes.POINTER(ctypes.c_uint8)]
_crc.crc_params.restype = params_t

_crc.crc_print_errors.argtypes = [ctypes.c_uint8]

_crc.crc_table.argtypes = [ctypes.POINTER(params_t), ctypes.c_uint64, ctypes.c_char_p, ctypes.c_uint64]
_crc.crc_table.restype = ctypes.c_uint64

_crc.crc_calc.argtypes = [ctypes.POINTER(params_t), ctypes.c_uint64, ctypes.c_char_p, ctypes.c_uint64]
_crc.crc_calc.restype = ctypes.c_uint64

_crc.crc_combine_constant.argtypes = [ctypes.POINTER(params_t), ctypes.c_uint64]
_crc.crc_combine_constant.restype = ctypes.c_uint64

_crc.crc_combine_fixed.argtypes = [ctypes.POINTER(params_t), ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64]
_crc.crc_combine_fixed.restype = ctypes.c_uint64

_crc.crc_combine.argtypes = [ctypes.POINTER(params_t), ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64]
_crc.crc_combine.restype = ctypes.c_uint64

cpu_check_features = _crc.cpu_check_features
cpu_enable_simd = ctypes.c_bool.in_dll(_crc, 'cpu_enable_simd')

def crc_params(width, poly, init, refin, refout, xorout, check):
    error = ctypes.c_uint8(0)
    params = _crc.crc_params(width, poly, init, refin, refout, xorout, check, ctypes.byref(error))

    if error.value > 0:
        _crc.crc_print_errors(error)
        raise ValueError('Invalid CRC parameters.')

    return params

def crc_table(params, crc, buf):
    return _crc.crc_table(ctypes.byref(params), crc, buf, len(buf))

def crc_calc(params, crc, buf):
    return _crc.crc_calc(ctypes.byref(params), crc, buf, len(buf))

def crc_calc_unaligned(params, crc, buf, shift):
    pointer = ctypes.cast(buf, ctypes.POINTER(ctypes.c_char))
    address = ctypes.addressof(pointer.contents)
    pointer2 = ctypes.cast(address + shift, ctypes.POINTER(ctypes.c_char))

    return _crc.crc_calc(ctypes.byref(params), crc, pointer2, len(buf) - shift)

def crc_combine_constant(params, len):
    return _crc.crc_combine_constant(ctypes.byref(params), len)

def crc_combine_fixed(params, crc, crc2, xp):
    return _crc.crc_combine_fixed(ctypes.byref(params), crc, crc2, xp)

def crc_combine(params, crc, crc2, len):
    return _crc.crc_combine(ctypes.byref(params), crc, crc2, len)


def get_hex(value, width):
    str_width = width // 4
    if width % 4 != 0:
        str_width += 1

    return ''.join(['0x{:0', str(str_width), 'x}']).format(value)

def str_model(model):
    return 'width={}, poly={}, init={}, refin={}, refout={}, xorout={}, check={}'.format(model.width, get_hex(model.poly, model.width),
    get_hex(model.init, model.width), model.refin, model.refout, get_hex(model.xorout, model.width), get_hex(model.check, model.width))


# This file is for experimenting with GF2 math and prototyping of CRC code
# Python makes this task easier with arbitrary length integers

# Everything is treated as a polynomial in the GF2 domain
# The coefficients of the polynomial are the bits (1s and 0s) of the integer
# Addition and subtraction are equivalent to an XOR
# Multiplication by x^n is equivalent to a left shift by n
# Division by x^n is equivalent to a right shift by n
# Multiplication and division of polynomials is carryless

from dataclasses import dataclass

def print_bin(a):
    print(bin(a))

def print_hex(a, w):
    print(('0x{:0' + str(w // 4) + 'x}').format(a))

@dataclass
class Model:
    width: int
    poly: int
    init: int
    refin: bool
    refout: bool
    xorout: int

crc32 = Model(32, 0x04c11db7, 0xffffffff, True, True, 0xffffffff)
crc32_mpeg = Model(32, 0x04c11db7, 0xffffffff, False, False, 0x00000000)
crc64_we = Model(64, 0x42f0e1eba9ea3693, 0xffffffffffffffff, False, False, 0xffffffffffffffff)
crc64_xz = Model(64, 0x42f0e1eba9ea3693, 0xffffffffffffffff, True, True, 0xffffffffffffffff)

# Find the remainder resulting from dividing polynomials a by b in GF2
# It tries to match the way this is done by hand
# You can find examples of this in the Ross Williams and Tad McCorkle guides
def mod(a, b):
    a_len = a.bit_length()
    b_len = b.bit_length()

    if a_len >= b_len:
        b <<= a_len - b_len

        for i in reversed(range(b_len - 1, a_len)):
            if (a >> i) & 1:
                a ^= b
            b >>= 1

    return a

# Expected: 1110
#print_bin(mod(0b11010110110000, 0b10011))

# Find the result of dividing polynomials a by b in GF2
# This is similar to mod but we apply a 1 every time the dividend is subtracted
def div(a, b):
    a_len = a.bit_length()
    b_len = b.bit_length()

    if a_len >= b_len:
        c = 0
        b <<= a_len - b_len

        for i in reversed(range(b_len - 1, a_len)):
            c <<= 1
            if (a >> i) & 1:
                a ^= b
                c |= 1
            b >>= 1

    return c

# Expected: 1100001010
#print_bin(div(0b11010110110000, 0b10011))

# Multiply polynomials a and b in GF2
def clmul(a, b):
    c = 0
    for _ in range(a.bit_length()):
        if a & 1:
            c ^= b
        a >>= 1
        b <<= 1

    return c

# Expected 1111111
#print_bin(clmul(0b1101, 0b1011))

#Reflect all the bits in an integer
def ref(a, w):
    b = 0
    for i in range(w):
        b <<= 1
        b |= a & 1
        a >>= 1

    return b

#Reflect all bytes in a bytes object
def ref_bytes(buf):
    return bytes((ref(b, 8) for b in buf))

# "Mathematical" CRC
# CRC is defined as M * X^W mod P
# M is the incoming data/message
# X^W is intentionally added so that all of the bits of M would be applied to the CRC
# otherwise polynomial division would stop before reaching the last W bits of M
# P is the polynomial including the often omitted X^(W + 1) term
# This also applies the different CRC parameters to the result
def crc_m(model, buf):
    # refin indicates that all of the bytes should be reflected before being consumed
    if model.refin:
        buf = ref_bytes(buf)

    data = int.from_bytes(buf, 'big')

    # The init is XOred with the first few incoming bits
    data ^= (model.init << (len(buf) * 8 - model.width))

    # Multiply by x^w
    data <<= model.width

    # Add x^(w+1)
    poly = (1 << model.width) | model.poly

    # Calculate the CRC
    crc = mod(data, poly)

    if model.refout:
        crc = ref(crc, model.width)
    return crc ^ model.xorout

# Prototype of the hardware accelerated algorithm
# Might have some unresolved errors
def crc_clmul(model, buf):
    if model.refin:
        p = (1 << model.width) | model.poly
        p <<= (64 - model.width)
        init = model.init

        k1 = ref(mod(1 << (512 + 63), p), 64)
        k2 = ref(mod(1 << (512 - 1), p), 64)

        x1 = x2 = x3 = x4 = 0
        h1 = h2 = h3 = h4 = 0
        l1 = l2 = l3 = l4 = 0

        x1 = int.from_bytes(buf[:16], 'little')
        x1 ^= init

        x2 = int.from_bytes(buf[16:32], 'little')
        x3 = int.from_bytes(buf[32:48], 'little')
        x4 = int.from_bytes(buf[48:64], 'little')

        length = len(buf) - 64
        pos = 64

        while length >= 64:
            h1 = clmul(x1 & 0xffffffffffffffff, k1)
            h2 = clmul(x2 & 0xffffffffffffffff, k1)
            h3 = clmul(x3 & 0xffffffffffffffff, k1)
            h4 = clmul(x4 & 0xffffffffffffffff, k1)

            l1 = clmul(x1 >> 64, k2)
            l2 = clmul(x2 >> 64, k2)
            l3 = clmul(x3 >> 64, k2)
            l4 = clmul(x4 >> 64, k2)

            x1 = int.from_bytes(buf[pos:pos + 16], 'little')
            x2 = int.from_bytes(buf[pos + 16:pos + 32], 'little')
            x3 = int.from_bytes(buf[pos + 32:pos + 48], 'little')
            x4 = int.from_bytes(buf[pos + 48:pos + 64], 'little')

            x1 ^= h1 ^ l1
            x2 ^= h2 ^ l2
            x3 ^= h3 ^ l3
            x4 ^= h4 ^ l4

            pos += 64
            length -= 64

        data = (x4 << 48 * 8) | (x3 << 32 * 8) | (x2 << 16 * 8) | x1
        data = ref(data, 64 * 8)
        data = (data << length * 8) | int.from_bytes(buf[pos:], 'big')
        crc = mod(data << 64, p) >> (64 - model.width)

        #return crc
        if model.refout:
            crc = ref(crc, model.width)
        return crc ^ model.xorout

    else:
        p = (1 << model.width) | model.poly
        p <<= (64 - model.width)
        init = model.init << (64 - model.width)

        k1 = mod(1 << (512 + 64), p)
        k2 = mod(1 << 512, p)

        x1 = x2 = x3 = x4 = 0
        h1 = h2 = h3 = h4 = 0
        l1 = l2 = l3 = l4 = 0

        x1 = int.from_bytes(buf[:16], 'big')
        x1 ^= (init << (128 - 64))

        x2 = int.from_bytes(buf[16:32], 'big')
        x3 = int.from_bytes(buf[32:48], 'big')
        x4 = int.from_bytes(buf[48:64], 'big')

        length = len(buf) - 64
        pos = 64

        while length >= 64:
            h1 = clmul(x1 >> 64, k1)
            h2 = clmul(x2 >> 64, k1)
            h3 = clmul(x3 >> 64, k1)
            h4 = clmul(x4 >> 64, k1)

            l1 = clmul(x1 & 0xffffffffffffffff, k2)
            l2 = clmul(x2 & 0xffffffffffffffff, k2)
            l3 = clmul(x3 & 0xffffffffffffffff, k2)
            l4 = clmul(x4 & 0xffffffffffffffff, k2)

            x1 = int.from_bytes(buf[pos:pos + 16], 'big')
            x2 = int.from_bytes(buf[pos + 16:pos + 32], 'big')
            x3 = int.from_bytes(buf[pos + 32:pos + 48], 'big')
            x4 = int.from_bytes(buf[pos + 48:pos + 64], 'big')

            x1 ^= h1 ^ l1
            x2 ^= h2 ^ l2
            x3 ^= h3 ^ l3
            x4 ^= h4 ^ l4

            pos += 64
            length -= 64

        data = (x1 << 48 * 8) | (x2 << 32 * 8) | (x3 << 16 * 8) | x4
        data = (data << length * 8) | int.from_bytes(buf[pos:], 'big')
        crc = mod(data << 64, p) >> (64 - model.width)

        #return crc
        if model.refout:
            crc = ref(crc, model.width)
        return crc ^ model.xorout

# Barret Reduction for 64-bit CRCs
# This is strictly following the Intel paper
def barret(model, value):
    p = (1 << model.width) | model.poly
    if model.refin:
        pp = ref(p, model.width + 1)
        u = ref(div(1 << 128, p), model.width + 1)
        t1 = clmul(value & 0xffffffffffffffff, u)
        t2 = clmul(t1 & 0xffffffffffffffff, pp)
        return (value ^ t2) >> 64
    else:
        u = div(1 << 128, p)
        t1 = clmul(value >> 64, u)
        t2 = clmul(t1 >> 64, p)
        return value ^ t2

# For testing the result of the Barret Reduction
#print_hex(mod(v, p), 64)

# Reflected case
#print_hex(ref(mod(v, p), 64), 64)


from setuptools import setup, Extension
import platform
import sys
import sysconfig

macros = []
use_limited_api = False
options = {}

#https://github.com/pypa/cibuildwheel/blob/main/cibuildwheel/architecture.py
if (platform.machine().lower() not in ('x86_64', 'amd64', 'aarch64', 'arm64', 'arm64_v8a')) or sys.maxsize < 2 ** 32:
    macros.append(('DISABLE_SIMD', None))

if sys.version_info.minor >= 11 and not sysconfig.get_config_var("Py_GIL_DISABLED"):
    use_limited_api = True
    macros.append(('Py_LIMITED_API', 0x030B0000))
    options = {'bdist_wheel': {'py_limited_api': 'cp311'}}

setup(
    ext_modules = [
        Extension(
            name='anycrc.crc',
            sources=['src/anycrc/crc.pyx',
                     'lib/crc-clmul/crc.c',
                     'lib/crc-clmul/cpu.c'],
            define_macros = macros,
            py_limited_api = use_limited_api
        )
    ],
    options = options
)


# Copyright (c) 2024-2026 Hussain Al Marzooq

# cython: freethreading_compatible = True

from libc.stdint cimport *
from functools import lru_cache
from bitarray import bitarray, frozenbitarray
from .models import models, aliases

cdef extern from '../../lib/crc-clmul/crc.h':
    ctypedef struct params_t:
        uint8_t width
        uint64_t poly
        bint refin, refout
        uint64_t init, xorout, k1, k2
        uint64_t table[256]
        uint64_t combine_table[64]

    cdef params_t crc_params(uint8_t width, uint64_t poly, uint64_t init, bint refin, bint refout, uint64_t xorout, uint64_t check, uint8_t *error)

    cdef uint64_t crc_table(params_t *params, uint64_t crc, const unsigned char *buf, uint64_t len)
    cdef uint64_t crc_calc(params_t *params, uint64_t crc, const unsigned char *buf, uint64_t len)
    cdef uint64_t crc_calc_bits(params_t *params, uint64_t crc, const unsigned char *buf, uint64_t len)

    cdef uint64_t crc_combine_constant(params_t *params, uint64_t len)
    cdef uint64_t crc_combine_constant_bits(params_t *params, uint64_t len)
    cdef uint64_t crc_combine_fixed(params_t *params, uint64_t crc, uint64_t crc2, uint64_t xp)

cdef class _CRC:
    cdef params_t params

    def __cinit__(self, width, poly, init, refin, refout, xorout, check=None):
        cdef uint8_t error
        self.params = crc_params(width, poly, init, refin, refout, xorout, check if check is not None else 0, &error)

        if error & 1:
            raise ValueError('CRC width should be between 1 and 64 bits')

        if check is not None and error & 0x20:
            raise ValueError('Invalid paramaters. check mismatch')

    def calc(self, data, init=None):
        if isinstance(data, bitarray) or isinstance(data, frozenbitarray):
            raise TypeError('Bitarray objects are not allowed, use calc_bits() instead')

        if isinstance(data, str):
            data = (<unicode> data).encode('utf-8')

        if init is None:
            init = self.params.init

        if len(data) == 0:
            return init

        cdef const unsigned char[:] view = data
        return crc_calc(&self.params, init, &view[0], len(view))

    def calc_bits(self, data, init=None):
        if not isinstance(data, bitarray) and not isinstance(data, frozenbitarray):
            raise TypeError('Expected a bitarray object')

        if self.params.refin and data.endian != 'little':
            raise ValueError('A little endian bitarray object is expected for reflected CRCs')

        if not self.params.refin and data.endian != 'big':
            raise ValueError('A big endian bitarray object is expected for non-reflected CRCs')

        if init is None:
            init = self.params.init

        if len(data) == 0:
            return init

        cdef const unsigned char[:] view = data
        return crc_calc_bits(&self.params, init, &view[0], len(data))

    @lru_cache
    def _combine_constant(self, length):
        return crc_combine_constant(&self.params, length)

    @lru_cache
    def _combine_constant_bits(self, length):
        return crc_combine_constant_bits(&self.params, length)

    def combine(self, crc1, crc2, length):
        return crc_combine_fixed(&self.params, crc1, crc2, self._combine_constant(length))

    def combine_bits(self, crc1, crc2, length):
        return crc_combine_fixed(&self.params, crc1, crc2, self._combine_constant_bits(length))

    #byte-by-byte (for testing)
    def _calc_b(self, data):
        cdef const unsigned char[:] view = data
        return crc_table(&self.params, self.params.init, &view[0], len(view))

@lru_cache
def CRC(width=None, poly=None, init=None, refin=None, refout=None, xorout=None, check=None):
    names = ('width', 'poly', 'init', 'refin', 'refout', 'xorout')
    values = (width, poly, init, refin, refout, xorout)

    for name, value in zip(names, values):
        if value is None:
            raise ValueError(f'{name} value is not provided')

    return _CRC(*values)

def Model(name):
    if name in models:
        return CRC(*models[name])
    elif name in aliases:
        return CRC(*models[aliases[name]])
    else:
        raise ValueError('CRC model not found')